/* X11 mixer
 * Copyright (C) 2000 timecop@japan.co.jp
 * User Interface Graphics (C) 1999 TC|Works GMBH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#include "master.xpm"
#include "mpg123.h"

static unsigned char joystick_mask[] = {
   0x3e, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x3e };

/* #define eprint(fmt, arg...) fprintf(stderr, fmt, ##arg) */
#define eprint(fmt, arg...) do { } while(0)
   

/* proto */
int add_mr(int x, int y, int width, int height);
int check_mr(int x, int y);
void do_joystick(int x, int y);

void keybd_event(GdkEventKey * ev);
void motion_event(GdkEventMotion * ev);
void press_event(GdkEventButton * ev);
void release_event(GdkEventButton * ev);

/* callback proto */
void on_off_cb(void);
void preset_cb(int nr, int btn);
void joystick_cb(int x, int y);
void sliders_cb(int slider, int y);

/* draw proto */
void draw_slider(int slider, int value);
void draw_joystick(int x, int y);
void draw_on_off(int state);
void draw_preset(int nr);

/* preset save/load functions */
void load_preset(int preset);
void save_preset(int preset);

typedef struct {
    int enable;
    int x;
    int y;
    int width;
    int height;
} MRegion;			/* Mouse region info */

/* globals */
Mixer *data;			/* "data" keeps all the info */
MRegion mr[32];			/* mouse region table */

int add_mr(int x, int y, int width, int height)
{
    static int index;
    index++;			/* start with 1 */
    mr[index].enable = 1;
    mr[index].x = x;
    mr[index].y = y;
    mr[index].width = width;
    mr[index].height = height;
    return index;
}

int check_mr(int x, int y)
{
    register int i;
    register int found = 0;

    for (i = 1; i < 32 && !found; i++) {
	if (mr[i].enable && x >= mr[i].x &&
	    x <= mr[i].x + mr[i].width &&
	    y >= mr[i].y && y <= mr[i].y + mr[i].height)
	    found = 1;
    }
    if (!found)
	return -1;
    return (i - 1);

}

void init_spline(float *x, float *y, int n, float *y2)
{
    int i, k;
    float p, qn, sig, un, *u;

    u = (float *) malloc(n * sizeof(float));

    y2[0] = u[0] = 0.0;

    for (i = 1; i < n - 1; i++) {
	sig = ((float) x[i] - x[i - 1]) / ((float) x[i + 1] - x[i - 1]);
	p = sig * y2[i - 1] + 2.0;
	y2[i] = (sig - 1.0) / p;
	u[i] =
	    (((float) y[i + 1] - y[i]) / (x[i + 1] - x[i])) -
	    (((float) y[i] - y[i - 1]) / (x[i] - x[i - 1]));
	u[i] = (6.0 * u[i] / (x[i + 1] - x[i - 1]) - sig * u[i - 1]) / p;
    }
    qn = un = 0.0;

    y2[n - 1] = (un - qn * u[n - 2]) / (qn * y2[n - 2] + 1.0);
    for (k = n - 2; k >= 0; k--)
	y2[k] = y2[k] * y2[k + 1] + u[k];
    free(u);
}

float eval_spline(float xa[], float ya[], float y2a[], int n, float x)
{
    int klo, khi, k;
    float h, b, a;

    klo = 0;
    khi = n - 1;
    while (khi - klo > 1) {
	k = (khi + klo) >> 1;
	if (xa[k] > x)
	    khi = k;
	else
	    klo = k;
    }
    h = xa[khi] - xa[klo];
    a = (xa[khi] - x) / h;
    b = (x - xa[klo]) / h;
    return (a * ya[klo] + b * ya[khi] +
	    ((a * a * a - a) * y2a[klo] +
	     (b * b * b - b) * y2a[khi]) * (h * h) / 6.0);
}

void eqgraph_draw(void)
{
    int i, y, ymin, ymax, py = 0;
    float x[] = { 0, 11, 23, 35, 47, 59, 71, 83, 97, 109, 121, 133, 145, 157, 169, 188}, yf[16];
    float slid[16];
    guchar rgb[189 * 58 * 3];

    for (i = 0; i < 16; i++)
	slid[i] = -(float) data->hw->slider[i];

    memcpy(rgb, pixel_data, sizeof(rgb));
    init_spline(x, slid, 16, yf);
    for (i = 0; i < 188; i++) {
	y = 28 - (int) ((eval_spline(x, slid, yf, 16, i) * 11.0) / 20.0);
	if (y < 0)
	    y = 0;
	if (y > 56)
	    y = 56;
	if (!i)
	    py = y;
	if (y < py) {
	    ymin = y;
	    ymax = py;
	} else {
	    ymin = py;
	    ymax = y;
	}
	py = y;
	for (y = ymin; y <= ymax; y++) {
	    rgb[y * 189 * 3 + (i * 3)] = 0x02;
	    rgb[y * 189 * 3 + (i * 3)+ 1] = 0xfe;
	    rgb[y * 189 * 3 + (i * 3)+ 2] = 0x02;
	    rgb[(y + 1) * 189 * 3 + (i * 3)] = 0x02;
	    rgb[(y + 1) * 189 * 3 + (i * 3)+ 1] = 0x82;
	    rgb[(y + 1) * 189 * 3 + (i * 3)+ 2] = 0x02;
	}
    }
    gdk_draw_rgb_image(data->ui->win, data->ui->gc, 234, 28, 189, 57, 
	    GDK_RGB_DITHER_NORMAL, rgb, 189 * 3);
}

void init_all(void)
{
    /* X11 data, to hack around GDK retardedness */
    XSetWindowAttributes xattr;
    XSizeHints sizehints;
    Window win;

    GdkWindowAttr attr;
    GdkImage *color;		/* eq window graph color */
    GdkColor pixel;		/* color pixel */
    GdkBitmap *mask;		/* joystick top mask */

    gdk_set_use_xshm(TRUE);	/* use SHM */

    attr.title = "mpg123eq";
    attr.event_mask = GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK |
	GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK |
	GDK_STRUCTURE_MASK | GDK_KEY_PRESS_MASK;

    attr.width = 557;
    attr.height = 340;
    attr.wclass = GDK_INPUT_OUTPUT;
    attr.visual = gdk_visual_get_system();
    attr.colormap = gdk_colormap_get_system();
    attr.wmclass_name = "mpg123eq";
    attr.wmclass_class = "mpg123eq";
    attr.window_type = GDK_WINDOW_TOPLEVEL;
    sizehints.min_width = attr.width;
    sizehints.min_height = attr.height;
    sizehints.max_width = attr.width;
    sizehints.max_height = attr.height;
    sizehints.flags = PMinSize | PMaxSize;
    xattr.save_under = True;
    xattr.backing_store = WhenMapped;
    /* make the window */

    data->ui->win = gdk_window_new(NULL, &attr,
				   GDK_WA_TITLE | GDK_WA_WMCLASS |
				   GDK_WA_VISUAL | GDK_WA_COLORMAP);

    /* make the background pixmap */
    data->ui->mixer = gdk_pixmap_create_from_xpm_d(data->ui->win,
						   NULL, NULL, master_xpm);

    gdk_window_set_back_pixmap(data->ui->win, data->ui->mixer, FALSE);

    /* make the window non-resizeable and use backing store */
    win = GDK_WINDOW_XWINDOW(data->ui->win);
    XSetWMNormalHints(GDK_WINDOW_XDISPLAY(data->ui->win), win, &sizehints);
    XChangeWindowAttributes(GDK_WINDOW_XDISPLAY(data->ui->win), win,
	    CWSaveUnder | CWBackingStore, &xattr);

    /* get images for UI */
    data->ui->button = gdk_image_get(data->ui->mixer, 0, 345, 73, 16);
    data->ui->handle = gdk_image_get(data->ui->mixer, 160, 345, 14, 29);
    data->ui->joystick = gdk_image_get(data->ui->mixer, 0, 369, 7, 7);
    data->ui->preset = gdk_image_get(data->ui->mixer, 0, 381, 57, 14);
    data->ui->eqbar = gdk_image_get(data->ui->mixer, 14, 362, 145, 18);

    data->ui->gc = gdk_gc_new(data->ui->win);
    data->ui->hig = gdk_gc_new(data->ui->win);
    data->ui->log = gdk_gc_new(data->ui->win);
    data->ui->joygc = gdk_gc_new(data->ui->win);

    /* setup high green and low green GC for equalizer graph drawing */
    color = gdk_image_get(data->ui->mixer, 0, 362, 2, 2);
    pixel.pixel = gdk_image_get_pixel(color, 0, 0);
    gdk_gc_set_foreground(data->ui->hig, &pixel);
    pixel.pixel = gdk_image_get_pixel(color, 0, 1);
    gdk_gc_set_foreground(data->ui->log, &pixel);

    /* setup joystick mask */
    mask = gdk_bitmap_create_from_data(data->ui->win, joystick_mask, 7, 7);
    gdk_gc_set_clip_mask(data->ui->joygc, mask);

    gdk_window_show(data->ui->win);
}

void do_joystick(int x, int y)
{
#ifdef UGLY
    int i;
    int val;
    int tmp;
    float ramp[16] =
	{ 5.4, 4.4, 3.5, 2.7, 2.0, 1.4, 1.2, 1.0, 1.0, 1.2, 1.4, 2.0, 2.7,
	    3.5, 4.4, 5.4 };

    /*
     * -x = more bass
     * x = less bass
     * -y = more treble
     * y = less treble
     */

    for (i = 0; i < 8; i++) {
	val = floor(ramp[i] * x);
	tmp = data->hw->slider[i] + val;
	if (tmp > 50)
	    tmp = 50;
	if (tmp < -50)
	    tmp = -50;
	eprint("slider: %d (delta %d)\n", i, val);
	draw_slider(i, tmp);
    }

    for (i = 8; i < 16; i++) {
	val = floor(ramp[i] * y);
	tmp = data->hw->slider[i] + val;
	if (tmp > 50)
	    tmp = 50;
	if (tmp < -50)
	    tmp = -50;
	eprint("slider: %d (delta %d)\n", i, val);
	draw_slider(i, tmp);
    }
#endif
}

void press_event(GdkEventButton * ev)
{
    int x = ev->x;
    int y = ev->y;
    int btn = ev->button;
    int region = check_mr(x, y);

    eprint("mouse click: %d, %d (%d)\n", x, y, btn);
    eprint("region: %d\n", region);

    switch (region) {
    case 1:
	eprint("pressed on/off button\n");
	on_off_cb();
	break;
    case 2:
    case 3:
    case 4:
	eprint("pressed one of preset buttons\n");
	preset_cb(region - 1, btn);
	break;
    case 5:
	eprint("pressed joystick\n");
	data->hw->joy_press = 1;
	joystick_cb(x, y);
	break;
    case 6:
	eprint("click inside slider area\n");
	data->hw->slider_press = 1;
	sliders_cb((x - 28) >> 5, y);	/* slider nr (0..15) */
	break;
    }
}

void release_event(GdkEventButton * ev)
{
    /* int x = ev->x;
    int y = ev->y;

    eprint("mouse release: %d, %d\n", x, y); */
    data->hw->joy_press = 0;
    data->hw->slider_press = 0;
}

void keybd_event(GdkEventKey * ev)
{
    eprint("key press: %d\n", ev->keyval);
}

void motion_event(GdkEventMotion * ev)
{
    int x = ev->x;
    int y = ev->y;

    if (data->hw->joy_press) {	/* joystick is in use */
	joystick_cb(x, y);
    }
    if (data->hw->slider_press) {	/* we are sliding something */
	sliders_cb(data->hw->currslider, y);
    }
    /* eprint("mouse move: %d, %d\n", x, y); */
}

/* callbacks */
void on_off_cb(void)
{
    data->enable = !data->enable;
    eprint("Button state: %d\n", data->enable);
    draw_on_off(data->enable);
}

void preset_cb(int nr, int btn)
{
    switch (btn) {
    case 1:			/* left button: select preset */
	data->preset = nr;
	draw_preset(data->preset);
	eprint("Loading preset %d\n", data->preset);
	load_preset(data->preset);
	break;
    case 2:			/* middle button: write preset */
	data->preset = nr;
	draw_preset(data->preset);
	eprint("Saving preset %d\n", data->preset);
	save_preset(data->preset);
	break;
    }
}

void load_preset(int preset)
{
    FILE *fp;
    int *v = data->hw->slider;
    int i;
    char buf[512];

    fp = fopen(data->config, "r");
    if (fp) {
	while (preset--)
	    fgets(buf, 512, fp);
	fclose(fp);
    }
    eprint("string: %s\n", buf);
    sscanf(buf, "%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d", &v[0],
	   &v[1], &v[2], &v[3], &v[4], &v[5], &v[6], &v[7], &v[8], &v[9],
	   &v[10], &v[11], &v[12], &v[13], &v[14], &v[15]);

    for (i = 0; i < 16; i++)
	draw_slider(i, data->hw->slider[i]);
    eqgraph_draw();

}

void save_preset(int preset)
{
    FILE *fp;
    int *v;
    int set[3][16];

    int i;
    char buf[512];
    fp = fopen(data->config, "r");
    if (fp) {
    for (i = 0; i < 3; i++) {
	fgets(buf, 512, fp);
	v = set[i];
	sscanf(buf, "%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d",
	       &v[0], &v[1], &v[2], &v[3], &v[4], &v[5], &v[6], &v[7],
	       &v[8], &v[9], &v[10], &v[11], &v[12], &v[13], &v[14],
	       &v[15]);
    }
    fclose(fp);
    } else {
	memset(set, 0, sizeof(int) * 3 * 16);
    }
    memcpy(set[preset - 1], data->hw->slider, sizeof(data->hw->slider));

    fp = fopen(data->config, "w");
    for (i = 0; i < 3; i++) {
	v = set[i];
	fprintf(fp, "%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
		v[0], v[1], v[2], v[3], v[4], v[5], v[6], v[7], v[8], v[9],
		v[10], v[11], v[12], v[13], v[14], v[15]);
    }
    fclose(fp);

}

void joystick_cb(int x, int y)
{
    int real_x = x;
    int real_y = y;
    float dist;

    real_x -= 474;
    real_y -= 56;
    real_y = -real_y;

    eprint("inside joystick, %2d, %2d\n", real_x, real_y);
    dist = sqrt((real_x * real_x) + (real_y * real_y));

    if (dist < 13.5) {
	draw_joystick(real_x, real_y);
	data->hw->joy_x = real_x;
	data->hw->joy_y = real_y;
	do_joystick(real_x, real_y);
    }
}

void sliders_cb(int slider, int y)
{
    int ypos;

    ypos = y - 14;
    if (ypos < 153)
	ypos = 153;
    if (ypos > 288)
	ypos = 288;

    ypos -= 153;
    ypos = (ypos * 100) / 135;
    ypos -= 50;

    eprint("we are on slider %d (%d)\n", slider, ypos);
    data->hw->currslider = slider;

    /* remember state */
    data->hw->slider[data->hw->currslider] = ypos;
	
    /* draw it */
    draw_slider(slider, ypos);
    eqgraph_draw();
}

/* draw functions just draw, without using any variables.  This is good for
 * a consistent draw interface, and for handling expose events (redrawing
 * all screen, without affecting data structures.  Also note, the sliders
 * here take the actual adjustment argument (-50 to 50) instead of a y axis
 * position */

void draw_slider(int slider, int value)
{
    int ypos;
    static int sliders[16] = { 0, };

    ypos += 50;			/* shift into positive scale */

    ypos = (value * 135) / 100;
    ypos += 221;
    eprint("slider: value: %d (old %d) y: %d\n", value, sliders[slider], ypos);

    /* set mpg123 equalizer */
    equalizer[0][slider] = (float) (100-(value+50)) / 50.0;
    equalizer[1][slider] = (float) (100-(value+50)) / 50.0;
    eprint("slider: %d, value: %f\n", slider, (100-(value+50)) / 50.0);

    gdk_window_clear_area(data->ui->win, (slider << 5) + 28, sliders[slider], 17, 29);
    gdk_draw_image(data->ui->win, data->ui->gc, data->ui->handle,
	    0, 0, (slider << 5) + 31, ypos, 14, 29);
    sliders[slider] = ypos;
}

void draw_joystick(int x, int y)
{
    static int oldx, oldy;
    x += 474;
    y = -y;
    y += 56;

    gdk_window_clear_area(data->ui->win, oldx - 3, oldy - 3, 7, 7);
    gdk_gc_set_clip_origin(data->ui->joygc, x - 3, y - 3);
    gdk_draw_image(data->ui->win, data->ui->joygc, data->ui->joystick,
	    0, 0, x - 3, y - 3, 7, 7);
    oldx = x;
    oldy = y;
}

void draw_preset(int nr)
{
    nr--;			/* decrement because we count from 0 here */
    gdk_window_clear_area(data->ui->win, 516, 44, 19, 54);
    gdk_draw_image(data->ui->win, data->ui->gc, data->ui->preset,
	    nr * 19, 0, 516, 44 + (nr * 20), 19, 14);
}

void draw_on_off(int state)
{
    /* stick a on or off image of the button on the screen */
    gdk_draw_image(data->ui->win, data->ui->gc, data->ui->button,
	    state * 37, 0, 459, 110, 35, 15);
}


void ui_loop(void)
{
    GdkEvent *event;
    int i;
    static int left, right;

/*    while (1) { */
	while (gdk_events_pending()) {
	    event = gdk_event_get();
	    if (event) {
		switch (event->type) {
		case GDK_EXPOSE:	/* expose */
		    draw_on_off(data->enable);
		    draw_preset(data->preset);
		    draw_joystick(data->hw->joy_x, data->hw->joy_y);	/* joystick (474, 56) */
		    for (i = 0; i < 16; i++)
			draw_slider(i, data->hw->slider[i]);
		    eqgraph_draw();
		    break;
		case GDK_DESTROY:
		    gdk_exit(0);
		    exit(0);
		    break;
		case GDK_BUTTON_PRESS:	/* button press */
		    press_event(&event->button);
		    break;
		case GDK_BUTTON_RELEASE:	/* button release */
		    release_event(&event->button);
		    break;
		case GDK_KEY_PRESS:	/* keyboard press */
		    keybd_event(&event->key);
		    break;
		case GDK_MOTION_NOTIFY:	/* mouse movement */
		    motion_event(&event->motion);
		    break;
		default:
		    eprint("mori\n");
		    break;
		}
	    }
	}
	eprint("left: %d right: %d\n", data->hw->eq_in_l, data->hw->eq_in_r);

	/* gdk_window_clear_area(data->ui->win, 41, 35, 145, 18); */
	gdk_window_clear_area(data->ui->win, 41 + left, 35, 146 - left, 7);
	gdk_window_clear_area(data->ui->win, 41 + right, 46, 146 - right, 7);
	gdk_draw_image(data->ui->win, data->ui->gc, data->ui->eqbar,
		0, 0, 41, 35, data->hw->eq_in_l, 7);
	gdk_draw_image(data->ui->win, data->ui->gc, data->ui->eqbar,
		0, 11, 41, 46, data->hw->eq_in_r, 7);

	left = data->hw->eq_in_l;
	right = data->hw->eq_in_r;

/*	usleep(10000);
    } */
}

void eq_setup(int argc, char **argv)
{
    char *tmp = NULL;
    
    data = calloc(1, sizeof(Mixer));
    data->ui = calloc(1, sizeof(Ui));
    data->hw = calloc(1, sizeof(Hw));
    tmp = getenv("HOME");
    if (tmp == NULL) {
	fprintf(stderr, "You are probably not on a UNIX system\n");
	exit(-1);
    }
    data->config = calloc(1, strlen(tmp) + 11);
    strcpy(data->config, tmp);
    strcat(data->config, "/.mpg123rc");
    eprint("conf file: %s\n", data->config);

    /* init GDK */
    if (!gdk_init_check(&argc, &argv))
	fprintf(stderr, "GDK init failed\n");
    gdk_rgb_init();

    /* setup display, main window, etc */
    init_all();
    /* setup mouse click regions for static controls */
    add_mr(459, 110, 35, 15);	/* (1) on-off toggle button */
    add_mr(516, 44, 19, 14);	/* (2) preset button 1 */
    add_mr(516, 64, 19, 14);	/* (3) preset button 2 */
    add_mr(516, 84, 19, 14);	/* (4) preset button 3 */
    add_mr(448, 30, 52, 52);	/* (5) joystick area */
    add_mr(28, 151, 499, 169);	/* (6) sliders area */

    data->preset = 1;		/* first preset loaded */
    data->enable = 1;		/* EQ is on */
    load_preset(data->preset);	/* load 1st preset */
}
